# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import datetime
import hashlib

import unittest2
from expects import equal, expect, raise_error

from apitools.base.py import encoding
from google.api.control import distribution, timestamp, metric_value, messages
from google.api.control import MetricKind


class TestUpdateHash(unittest2.TestCase):
    NOTHING_ADDED = hashlib.md5().digest()

    def make_hash(self, mv):
        md5 = hashlib.md5()
        metric_value.update_hash(md5, mv)
        return md5.digest()

    def test_should_add_nothing_without_labels_or_currency(self):
        expect(self.make_hash(messages.MetricValue())).to(
            equal(self.NOTHING_ADDED))

    def test_should_add_matching_hashes_for_matching_labels(self):
        a_dict = {'test': 'dict'}
        mv1 = metric_value.create(labels=a_dict)
        mv2 = metric_value.create(labels=a_dict)
        want = self.make_hash(mv1)
        got = self.make_hash(mv2)
        expect(got).to(equal(want))

    def test_should_update_hash_for_when_currency_is_added(self):
        a_dict = {'test': 'dict'}
        mv1 = metric_value.create(labels=a_dict)
        mv2 = metric_value.create(labels=a_dict)
        mv2.moneyValue = messages.Money(currencyCode='JPY')
        want = self.make_hash(mv1)
        got = self.make_hash(mv2)
        expect(got).to_not(equal(want))


class TestSign(TestUpdateHash):

    def make_hash(self, mv):
        return metric_value.sign(mv)


class TestMerge(unittest2.TestCase):
    A_FLOAT_VALUE = 1.0
    EARLY = timestamp.to_rfc3339(datetime.datetime(1970, 1, 1, 10, 0, 0))
    LATER = timestamp.to_rfc3339(datetime.datetime(1990, 1, 1, 10, 0, 0))
    TEST_LABELS = {
        'key1': 'value1',
        'key2': 'value2',
    }

    def setUp(self):
        self.test_value = metric_value.create(
            labels=self.TEST_LABELS,
            doubleValue=self.A_FLOAT_VALUE)
        self.early_ending = metric_value.create(
            labels=self.TEST_LABELS,
            doubleValue=self.A_FLOAT_VALUE,
            endTime=self.EARLY)
        self.late_ending = metric_value.create(
            labels=self.TEST_LABELS,
            doubleValue=self.A_FLOAT_VALUE,
            endTime=self.LATER)
        self.test_value_with_money = metric_value.create(
            labels=self.TEST_LABELS,
            moneyValue=messages.Money(currencyCode='JPY', units=100, nanos=0))

    def test_should_fail_for_metric_values_with_different_types(self):
        changed = metric_value.create(labels=self.TEST_LABELS, int64Value=1)
        for kind in (MetricKind.GAUGE, MetricKind.CUMULATIVE, MetricKind.DELTA):
            testf = lambda: metric_value.merge(kind, self.test_value, changed)
            expect(testf).to(raise_error(ValueError))

    def test_should_fail_for_uninitialized_metric_values(self):
        no_init = metric_value.create()
        for kind in (MetricKind.GAUGE, MetricKind.CUMULATIVE, MetricKind.DELTA):
            testf = lambda: metric_value.merge(kind, no_init, no_init)
            expect(testf).to(raise_error(ValueError))

    def test_should_fail_for_delta_metrics_with_unmergable_types(self):
        no_init = metric_value.create()
        unmergeables = [
            metric_value.create(stringValue='a test string'),
            metric_value.create(boolValue=False),
        ]
        for mv in unmergeables:
            testf = lambda: metric_value.merge(MetricKind.DELTA, mv, mv)
            expect(testf).to(raise_error(ValueError))

    def test_should_succeed_for_delta_metrics_with_the_money_type(self):
        v = self.test_value_with_money
        want = 2 * v.moneyValue.units
        got = metric_value.merge(MetricKind.DELTA, v, v)
        expect(got.moneyValue.units).to(equal(want))

    def test_should_succeed_for_delta_metrics_with_the_double_type(self):
        v = self.test_value
        want = 2 * v.doubleValue
        got = metric_value.merge(MetricKind.DELTA, v, v)
        expect(got.doubleValue).to(equal(want))

    def test_should_succeed_for_delta_metrics_with_the_int64_type(self):
        test_int = 4
        v = metric_value.create(labels=self.TEST_LABELS, int64Value=test_int)
        want = 2 * test_int
        got = metric_value.merge(MetricKind.DELTA, v, v)
        expect(got.int64Value).to(equal(want))

    def test_should_succeed_for_delta_metrics_with_the_distribution_type(self):
        test_distribution = distribution.create_explicit([0.1, 0.3, 0.5])
        distribution.add_sample(0.4, test_distribution)
        v = metric_value.create(labels=self.TEST_LABELS,
                                distributionValue=test_distribution)
        want = 2 * test_distribution.count
        got = metric_value.merge(MetricKind.DELTA, v, v)
        expect(got.distributionValue.count).to(equal(want))

    def test_should_return_metric_value_with_latest_end_time_for_non_deltas(self):
        for kind in (MetricKind.GAUGE, MetricKind.CUMULATIVE):
            got = metric_value.merge(kind, self.early_ending, self.late_ending)
            expect(got).to(equal(self.late_ending))
            got = metric_value.merge(kind, self.late_ending, self.early_ending)
            expect(got).to(equal(self.late_ending))

    def test_should_use_the_latest_end_time_delta_merges(self):
        got = metric_value.merge(MetricKind.DELTA,
                                 self.early_ending,
                                 self.late_ending)
        expect(got.endTime).to(equal(self.late_ending.endTime))
        got = metric_value.merge(MetricKind.DELTA,
                                 self.late_ending,
                                 self.early_ending)
        expect(got.endTime).to(equal(self.late_ending.endTime))

    def test_should_use_the_earliest_start_time_in_delta_merges(self):
        early_starting = metric_value.create(
            labels=self.TEST_LABELS,
            doubleValue=self.A_FLOAT_VALUE,
            startTime=self.EARLY)
        late_starting = metric_value.create(
            labels=self.TEST_LABELS,
            doubleValue=self.A_FLOAT_VALUE,
            startTime=self.LATER)
        got = metric_value.merge(MetricKind.DELTA, early_starting,
                                 late_starting)
        expect(got.startTime).to(equal(early_starting.startTime))
        got = metric_value.merge(MetricKind.DELTA, late_starting,
                                 early_starting)
        expect(got.startTime).to(equal(early_starting.startTime))



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides a method for fetching Service Configuration from Google Service
Management API."""


import logging
import json
import os
import urllib3

from apitools.base.py import encoding
import google.api.gen.servicecontrol_v1_messages as messages
from oauth2client import client
from urllib3.contrib import appengine


logger = logging.getLogger(__name__)

_GOOGLE_API_SCOPE = "https://www.googleapis.com/auth/cloud-platform"

_SERVICE_MGMT_URL_TEMPLATE = ("https://servicemanagement.googleapis.com"
                              "/v1/services/{}/config?configId={}")

_SERVICE_NAME_ENV_KEY = "ENDPOINTS_SERVICE_NAME"
_SERVICE_VERSION_ENV_KEY = "ENDPOINTS_SERVICE_VERSION"


def fetch_service_config(service_name=None, service_version=None):
  """Fetches the service config from Google Serivce Management API.

  Args:
    service_name: the service name. When this argument is unspecified, this
      method uses the value of the "SERVICE_NAME" environment variable as the
      service name, and raises ValueError if the environment variable is unset.
    service_version: the service version. When this argument is unspecified,
      this method uses the value of the "SERVICE_VERSION" environment variable
      as the service version, and raises ValueError if the environment variable
      is unset.

  Returns: the fetched service config JSON object.

  Raises:
    ValueError: when the service name/version is neither provided as an
      argument or set as an environment variable; or when the fetched service
      config fails validation.
    Exception: when the Google Service Management API returns non-200 response.
  """
  if not service_name:
    service_name = _get_env_var_or_raise(_SERVICE_NAME_ENV_KEY)
  if not service_version:
    service_version = _get_env_var_or_raise(_SERVICE_VERSION_ENV_KEY)

  service_mgmt_url = _SERVICE_MGMT_URL_TEMPLATE.format(service_name,
                                                       service_version)

  access_token = _get_access_token()
  headers = {"Authorization": "Bearer {}".format(access_token)}

  http_client = _get_http_client()
  response = http_client.request("GET", service_mgmt_url, headers=headers)

  status_code = response.status
  if status_code != 200:
    message_template = "Fetching service config failed (status code {})"
    _log_and_raise(Exception, message_template.format(status_code))

  logger.debug('obtained service json from the management api:\n%s', response.data)
  service = encoding.JsonToMessage(messages.Service, response.data)
  _validate_service_config(service, service_name, service_version)
  return service


def _get_access_token():
  credentials = client.GoogleCredentials.get_application_default()
  if credentials.create_scoped_required():
    credentials = credentials.create_scoped(_GOOGLE_API_SCOPE)
  return credentials.get_access_token().access_token


def _get_http_client():
  if appengine.is_appengine_sandbox():
    return appengine.AppEngineManager()
  else:
    return urllib3.PoolManager()


def _get_env_var_or_raise(env_variable_name):
  if env_variable_name not in os.environ:
    message_template = 'The "{}" environment variable is not set'
    _log_and_raise(ValueError, message_template.format(env_variable_name))
  return os.environ[env_variable_name]


def _validate_service_config(service, expected_service_name,
                             expected_service_version):
  service_name = service.name
  if not service_name:
    _log_and_raise(ValueError, "No service name in the service config")
  if service_name != expected_service_name:
    message_template = "Unexpected service name in service config: {}"
    _log_and_raise(ValueError, message_template.format(service_name))

  service_version = service.id
  if not service_version:
    _log_and_raise(ValueError, "No service version in the service config")
  if service_version != expected_service_version:
    message_template = "Unexpected service version in service config: {}"
    _log_and_raise(ValueError, message_template.format(service_version))


def _log_and_raise(exception_class, message):
  logger.error(message)
  raise exception_class(message)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generated client library for servicecontrol version v1."""

# NOTE: This file is originally auto-generated using google-apitools then
# style-correcting hand edits were applied.  New behaviour should not provided
# by hand, please re-generate and restyle.
from __future__ import absolute_import

from apitools.base.py import base_api
from . import servicecontrol_v1_messages as messages


class ServicecontrolV1(base_api.BaseApiClient):
    """Generated client library for service servicecontrol version v1."""

    MESSAGES_MODULE = messages

    _PACKAGE = u'servicecontrol'
    _SCOPES = [u'https://www.googleapis.com/auth/cloud-platform',
               u'https://www.googleapis.com/auth/servicecontrol']
    _VERSION = u'v1'
    _CLIENT_CLASS_NAME = u'ServicecontrolV1'
    _URL_VERSION = u'v1'
    _API_KEY = None

    # pylint: disable=too-many-arguments
    def __init__(self, url='', credentials=None,
                 get_credentials=True, http=None, model=None,
                 log_request=False, log_response=False,
                 credentials_args=None, default_global_params=None,
                 additional_http_headers=None):
        """Create a new servicecontrol handle."""
        url = url or u'https://servicecontrol.googleapis.com/'
        super(ServicecontrolV1, self).__init__(
            url, credentials=credentials,
            get_credentials=get_credentials, http=http, model=model,
            log_request=log_request, log_response=log_response,
            credentials_args=credentials_args,
            default_global_params=default_global_params,
            additional_http_headers=additional_http_headers)
        self.services = self.ServicesService(self)

    class ServicesService(base_api.BaseApiService):
        """Service class for the services resource."""

        _NAME = u'services'

        def __init__(self, client):
            super(ServicecontrolV1.ServicesService, self).__init__(client)
            self._method_configs = {
                'check': base_api.ApiMethodInfo(
                    http_method=u'POST',
                    method_id=u'servicecontrol.services.check',
                    ordered_params=[u'serviceName'],
                    path_params=[u'serviceName'],
                    query_params=[],
                    relative_path=u'v1/services/{serviceName}:check',
                    request_field=u'checkRequest',
                    request_type_name=u'ServicecontrolServicesCheckRequest',
                    response_type_name=u'CheckResponse',
                    supports_download=False,
                ),
                'report': base_api.ApiMethodInfo(
                    http_method=u'POST',
                    method_id=u'servicecontrol.services.report',
                    ordered_params=[u'serviceName'],
                    path_params=[u'serviceName'],
                    query_params=[],
                    relative_path=u'v1/services/{serviceName}:report',
                    request_field=u'reportRequest',
                    request_type_name=u'ServicecontrolServicesReportRequest',
                    response_type_name=u'ReportResponse',
                    supports_download=False,
                ),
            }

            self._upload_configs = {
            }

        def check(self, request, global_params=None):
            """Checks quota, abuse status etc. to decide whether the given
            operation. should proceed. It should be called by the service
            before the given operation is executed.
            This method requires the `servicemanagement.services.check`
            permission on the specified service. For more information, see
            [Google Cloud IAM](https://cloud.google.com/iam).
            Args:
              request: (ServicecontrolServicesCheckRequest) input message
              global_params: (StandardQueryParameters, default: None)
                global arguments
            Returns:
              (CheckResponse) The response message.
            """
            config = self.GetMethodConfig('check')
            return self._RunMethod(
                config, request, global_params=global_params)

        def report(self, request, global_params=None):
            """Reports an operation to the service control features such as
            billing, logging, monitoring etc. It should be called by the
            service after the given operation is completed.
            This method requires the `servicemanagement.services.report`
            permission on the specified service. For more information, see
            [Google Cloud IAM](https://cloud.google.com/iam).
            Args:
              request: (ServicecontrolServicesReportRequest) input message
              global_params: (StandardQueryParameters, default: None) global
                arguments
            Returns:
              (ReportResponse) The response message.
            """
            config = self.GetMethodConfig('report')
            return self._RunMethod(
                config, request, global_params=global_params)



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines several suppliers that are used by the authenticator."""

import datetime
from dogpile import cache
from jwkest import jwk
import requests
import ssl


_HTTP_PROTOCOL_PREFIX = "http://"
_HTTPS_PROTOCOL_PREFIX = "https://"

_OPEN_ID_CONFIG_PATH = ".well-known/openid-configuration"


class KeyUriSupplier(object):  # pylint: disable=too-few-public-methods
  """A supplier that provides the `jwks_uri` for an issuer."""

  def __init__(self, issuer_uri_configs):
    """Construct an instance of KeyUriSupplier.

    Args:
      issuer_uri_configs: a dictionary mapping from an issuer to its jwks_uri
        configuration.
    """
    self._issuer_uri_configs = issuer_uri_configs

  def supply(self, issuer):
    """Supplies the `jwks_uri` for the given issuer.

    Args:
      issuer: the issuer.

    Returns:
      The `jwks_uri` that is either statically configured or retrieved via
      OpenId discovery. None is returned when the issuer is unknown or the
      OpenId discovery fails.
    """
    issuer_uri_config = self._issuer_uri_configs.get(issuer)

    if not issuer_uri_config:
      # The issuer is unknown.
      return

    jwks_uri = issuer_uri_config.jwks_uri
    if jwks_uri:
      # When jwks_uri is set, return it directly.
      return jwks_uri

    # When jwksUri is empty, we try to retrieve it through the OpenID
    # discovery.
    open_id_valid = issuer_uri_config.open_id_valid
    if open_id_valid:
      discovered_jwks_uri = _discover_jwks_uri(issuer)
      self._issuer_uri_configs[issuer] = IssuerUriConfig(False,
                                                         discovered_jwks_uri)
      return discovered_jwks_uri


class JwksSupplier(object):  # pylint: disable=too-few-public-methods
  """A supplier that returns the Json Web Token Set of an issuer."""

  def __init__(self, key_uri_supplier):
    """Constructs an instance of JwksSupplier.

    Args:
      key_uri_supplier: a KeyUriSupplier instance that returns the `jwks_uri`
        based on the given issuer.
    """
    self._key_uri_supplier = key_uri_supplier
    self._jwks_cache = cache.make_region().configure(
        "dogpile.cache.memory", expiration_time=datetime.timedelta(minutes=5))

  def supply(self, issuer):
    """Supplies the `Json Web Key Set` for the given issuer.

    Args:
      issuer: the issuer.

    Returns:
      The successfully retrieved Json Web Key Set. None is returned if the
        issuer is unknown or the retrieval process fails.

    Raises:
      UnauthenticatedException: When this method cannot supply JWKS for the
        given issuer (e.g. unknown issuer, HTTP request error).
    """
    def _retrieve_jwks():
      """Retrieve the JWKS from the given jwks_uri when cache misses."""
      jwks_uri = self._key_uri_supplier.supply(issuer)

      if not jwks_uri:
        raise UnauthenticatedException("Cannot find the `jwks_uri` for issuer "
                                       "%s: either the issuer is unknown or "
                                       "the OpenID discovery failed" % issuer)

      try:
        response = requests.get(jwks_uri)
        json_response = response.json()
      except Exception as exception:
        message = "Cannot retrieve valid verification keys from the `jwks_uri`"
        raise UnauthenticatedException(message, exception)

      if "keys" in json_response:
        # De-serialize the JSON as a JWKS object.
        jwks_keys = jwk.KEYS()
        jwks_keys.load_jwks(response.text)
        return jwks_keys._keys
      else:
        # The JSON is a dictionary mapping from key id to X.509 certificates.
        # Thus we extract the public key from the X.509 certificates and
        # construct a JWKS object.
        return _extract_x509_certificates(json_response)

    return self._jwks_cache.get_or_create(issuer, _retrieve_jwks)


def _extract_x509_certificates(x509_certificates):
  keys = []
  for kid, certificate in x509_certificates.iteritems():
    try:
      if certificate.startswith(jwk.PREFIX):
        # The certificate is PEM-encoded
        der = ssl.PEM_cert_to_DER_cert(certificate)
        key = jwk.der2rsa(der)
      else:
        key = jwk.import_rsa_key(certificate)
    except Exception as exception:
      raise UnauthenticatedException("Cannot load X.509 certificate",
                                     exception)
    rsa_key = jwk.RSAKey().load_key(key)
    rsa_key.kid = kid
    keys.append(rsa_key)
  return keys


def _discover_jwks_uri(issuer):
  open_id_url = _construct_open_id_url(issuer)
  try:
    response = requests.get(open_id_url)
    return response.json().get("jwks_uri")
  except Exception as error:
    raise UnauthenticatedException("Cannot discover the jwks uri", error)


def _construct_open_id_url(issuer):
  url = issuer
  if (not url.startswith(_HTTP_PROTOCOL_PREFIX) and
      not url.startswith(_HTTPS_PROTOCOL_PREFIX)):
    url = _HTTPS_PROTOCOL_PREFIX + url
  if not url.endswith("/"):
    url += "/"
  url += _OPEN_ID_CONFIG_PATH
  return url


class IssuerUriConfig(object):
  """The jwks_uri configuration for an issuer.

  TODO (yangguan): this class should be removed after we figure out how to
  fetch the external configs.
  """

  def __init__(self, open_id_valid, jwks_uri):
    """Create an instance of IsserUriConfig.

    Args:
      open_id_valid: indicates whether the corresponding issuer is valid for
        OpenId discovery.
      jwks_uri: is the saved jwks_uri. Its value can be None if the OpenId
        discovery process has not begun or has already failed.
    """
    self._open_id_valid = open_id_valid
    self._jwks_uri = jwks_uri

  @property
  def open_id_valid(self):
    return self._open_id_valid

  @property
  def jwks_uri(self):
    return self._jwks_uri


class UnauthenticatedException(Exception):
  pass
